
/*
Part of Tims Hybrid Harmonic Drive Project. https://www.instructables.com/Tims-Hybrid-Harmonic-Drive/

	By Tim Jackson
	Use Freely, Just give me credit where due :)

	NOTE!
	Useing an ESP8266
	Use Arduino "Board Manager" to install files for the ESP8266.

	Credits:
		Arduino.
		Espressif Systems.

	The Motor Driver used is a L298, This driver can brake when both INs are HIGH.
	Inputs:
		H/L=Forward.
		L/H=Reverse.
		Both same = Stop. (Enable HIGH)
		Enable LOW = Free Running.

	My work is always, work in progress.

	More projects here:
	https://timsblogplace.blogspot.com/2018/07/blog-post.html

	Current pins being used:

		D0 =	GPIO16	=	16	Wake up
		D1 =	GPIO5	=	5	SCL
		D2 =	GPIO4	=	4	SDA
		D3 =	GPIO0	=	0	MOTOR_FORWARD
		D4 =	GPIO2	=	2	MOTOR_REVERSE
		D5 =	GPIO14	=	14	MOTOR_SPEED
		D6 =	GPIO12	=	12	SENS_A
		D7 =	GPIO13	=	13	SENS_B
		D8 =	GPIO15	=	15
		RX =	GPIO3	=	3	RX
		TX =	GPIO1	=	1	TX
		AO =	ADC0	=	A0

	Interupts on 12 and 13.

		Command letters used:

		Letter		Device		Use
		H			Gear		Position Harmonic Gear.

	ESP8266 NodeMCU 4M (2M SPIFFS)
		Sketch uses 332624 bytes (31%) of program storage space. Maximum is 1044464 bytes.
		Global variables use 37848 bytes (46%) of dynamic memory, leaving 44072 bytes for local variables. Maximum is 81920 bytes.


*/
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include "FS.h"

#include "Page_by_val_HTML.h"
//#include "Page_HTML.h"

//#define DEBUG
/*
	Type of connection?
	"Access Point" or "Access Point with password"

		"Access Point"
		This is not connected to your local network, you connect to the Network of the ESP32-CAM.
		This means you can connect to it with your Moble Phone anywhare you are.
		You change the WiFi your Phone is connected to, to the ESP32-CAM WiFi.
		Then open browser to the IP of the ESP32-CAM control page. (may be 192.168.4.1)
		Have your serial monitor connected when you re-set the module to confirm correct IP Address.

		"Access Point Pasword"
		This is same as above, bvut you need a password to connect to the WiFi. (it currently is: 2468)

		Un-Comment which way you want to connect.
*/
//#define ACCESS_POINT
#define ACCESS_POINT_PW
/*
	The name of network of the ESP32-CAM
	The Password if used

	Note!
		If changing password, make sure it is at least 8 charectors long.
		If the password is too short, it will not be updated.
		Also:
			To change WiFi Setings, make sure that the "Erase Flash" setting is set correctly in "Tools" of the Arduino IDE.
			Only use "all" if you want to remove also files in SPIFFS.
*/
const char* ESP8266_SSID_AP = "ESP8266 WiFi Drive";
const char* ESP8266_PASSWORD_AP = "12345678";

#define MAX_BUF	64
#define SENS_A	12
#define SENS_B	13
#define PINGS_PER_REV	(4.0 * 4.0 * 1240.0)  // = 19,840.0  360/19840 = 0.0181451612903226 degrees
#define MOTOR_SPEED		14
#define MOTOR_FORWARD	0
#define MOTOR_REVERSE	2
#define MOTOR_SLOWDOWN	300

volatile byte seqA = 0;
volatile byte seqB = 0;
volatile byte cnt1 = 0;
volatile byte cnt2 = 0;
volatile long NewPossition = 0;//	New			position.
volatile long ReqPossition = 0;//	Requested	position.
volatile long CurPossition = 0;//	Current		position.
volatile bool StopMotor = false;
volatile bool SlowMotor = false;

volatile float Angle = 0;
volatile unsigned int Current_Speed = 1023;
volatile unsigned int Request_Speed = 1023;

/*
	SERIAL MESSAGES
*/
char buffer[MAX_BUF];			// Serial message.
unsigned int no_data = 0;
unsigned int Data_Count;		// Size of Serial Buffer.
bool isComment = false;
bool ProcsessingCommand = false;
short cmd = -1;
bool slaveProcesing = false;

bool Toggle = true;

/*
	Server
*/
IPAddress local_ip(192, 168, 4, 1);           //IP for AP mode
IPAddress gateway(192, 168, 100, 1);          //IP for AP mode
IPAddress subnet(255, 255, 255, 0);           //IP for AP mode
ESP8266WebServer server(80);

bool spiffsActive = false;
#define IMAGE_FILE "/Front_View_360x300.png"

/*	=====	Interupts	=====	*/

/*
	To do when there is an interupt.
*/
void PinChange() {

	/*
		Read A and B signals
	*/
	boolean A_val = digitalRead(SENS_A);
	boolean B_val = digitalRead(SENS_B);

	/*
		Record the A and B signals in seperate sequences
	*/
	seqA <<= 1;
	seqA |= A_val;

	seqB <<= 1;
	seqB |= B_val;

	/*
		Mask the MSB four bits
	*/
	seqA &= 0b00001111;
	seqB &= 0b00001111;

	/*
		Compare the recorded sequence with the expected sequence
	*/
	if (seqA == 0b00001001 && seqB == 0b00000011) {
		NewPossition++;
		if (NewPossition >= PINGS_PER_REV) {
			NewPossition = 0;
		}
		if (NewPossition >= ReqPossition) {
			StopMotor = true;
		}
		if (NewPossition >= (ReqPossition - MOTOR_SLOWDOWN)) {
			SlowMotor = true;
		}
	}

	if (seqA == 0b00000011 && seqB == 0b00001001) {
		NewPossition--;
		if (NewPossition <= -PINGS_PER_REV) {
			NewPossition = 0;
		}
		if (NewPossition <= ReqPossition) {
			StopMotor = true;
		}
		if (NewPossition <= (ReqPossition + MOTOR_SLOWDOWN)) {
			SlowMotor = true;
		}
	}
	CurPossition = NewPossition;

}

/*	=====	Motor	=====	*/

/*
	Set the speed of the motor.
	Sppeed 0 to 1024
*/
void SetSpeed(unsigned int speed) {
	Current_Speed = speed;
	if (speed == 1023) {
		analogWrite(MOTOR_SPEED, 0);
		digitalWrite(MOTOR_SPEED, HIGH);
	}
	else if (speed == 0) {
		analogWrite(MOTOR_SPEED, 0);
	}
	else {
		analogWrite(MOTOR_SPEED, speed);
	}
}
/*
	Release power to the motor.
*/
void MotorCoast() {

	digitalWrite(MOTOR_FORWARD, LOW);
	digitalWrite(MOTOR_REVERSE, LOW);
	digitalWrite(MOTOR_SPEED, LOW);

}
/*
	Turn on Brake.
*/
void MotorBrake() {

	digitalWrite(MOTOR_FORWARD, HIGH);
	digitalWrite(MOTOR_REVERSE, HIGH);
	digitalWrite(MOTOR_SPEED, HIGH);

}
/*
	Move to a position.
		Angle 0 to 359
		Clockwise true or false
		Current_Speed 0 to 1023
*/
void MoveGear(float angle, bool clockwise, unsigned int speed) {

	if (angle >= 360 || angle <= -360) { angle = 0; }

	ReqPossition = (float)(PINGS_PER_REV / 360.0) * angle;
	if (!clockwise) { ReqPossition *= -1; }

	MotorCoast();
	delay(10);
	SlowMotor = false;
	SetSpeed(speed);
	if (ReqPossition != CurPossition) {
		if (ReqPossition > CurPossition) { digitalWrite(MOTOR_FORWARD, HIGH); }
		else { digitalWrite(MOTOR_REVERSE, HIGH); }
	}
}
/*
	Perform stop sequece.
*/
void MotorStop() {

	MotorBrake();
	delay(500);
	MotorCoast();

	//#ifdef DEBUG
	Serial.print("NewPossition ");
	Serial.println(NewPossition);
	Serial.print("Angle: ");
	Serial.print((float)360.0 / float(PINGS_PER_REV / NewPossition), 4);
	Serial.println(" deg.");

	////#endif // DEBUG
	//delay(200);
	//server.send(200, "text/html", PAGE_HTML);


}
/*
	Back and forth test.
*/
void Do_Test() {

	MoveGear(45, Toggle, 255);
	Toggle = !Toggle;
	delay(10000);
}

/*	=====	Server	=====	*/

/*
	HTTP request: On Connect
*/
void Handle_Root() {
	Serial.println("Client connected");
	server.send(200, "text/html", PAGE_HTML);
}
/*
	HTTP request: Forward
*/
void Handle_0() {
	MoveGear(0, true, Request_Speed);
	Serial.println("Angle: 0");
	server.send(200, "text/html", PAGE_HTML);
}
/*
	HTTP request: Reverse
*/
void Handle_90() {
	MoveGear(90, true, Request_Speed);
	Serial.println("Angle: 90");
	server.send(200, "text/html", PAGE_HTML);
}
/*
	HTTP request: Stop.
*/
void Handle_180() {
	MoveGear(180, true, Request_Speed);
	Serial.println("Angle: 180");
	server.send(200, "text/html", PAGE_HTML);
}
/*
	HTTP request: 404
*/
void Handle_Not_Found() {
	Serial.println("Error 404 not found");
	server.send(404, "text/plain", "Not found");
}
/*
	Read Arguments received.
*/
void Handle_Position() {
	if (server.arg("value") == "stop") {
		MotorStop();
		Serial.println("Motor Stoped");
	}
	else {
		int thisAngle = server.arg("value").toInt();
		MoveGear(thisAngle, true, Request_Speed);
		Serial.print("Angle: ");
		Serial.println(thisAngle);
	}
	server.send(200, "text/html", PAGE_HTML);
}
/*
	Set Current_Speed.
*/
void Handle_Set_Speed() {
	Request_Speed = server.arg("value").toInt();
	SetSpeed(Request_Speed);
	Serial.print("Current_Speed: ");
	Serial.println(Current_Speed);
	server.send(200, "text/html", PAGE_HTML);
}
/*
	Send Image.
		Open it.
		send it to the client.
		close the file.
*/
void Handle_Front_View_IMAGE() {

	Serial.println("image request");
	if (spiffsActive) {
		File file = SPIFFS.open(IMAGE_FILE, "r");
		size_t sent = server.streamFile(file, "image/png");
		file.close();
	}
	else {
		Serial.println("SPIFFS is not active.\nImage cold notbe loaded.");
	}
}
/*
	Send Current Speed
*/
void Handle_Get_Speed() {

	String Speed_String = String(Request_Speed);
	server.send(200, "text/plane", Speed_String);
}
/*
	Send current Position
*/
void Handle_Get_Position() {

	String NewPossition_String = String((float)360.0 / float(PINGS_PER_REV / NewPossition));
	server.send(200, "text/plane", "True Position = " + String(NewPossition) + " Ticks = " + NewPossition_String + "&deg;");
}

/*	=====	Serial	=====	*/

/*
	Change value to a float.
*/
float ParseNumber(char code, float val) {

	char* ptr = buffer;															// start at the beginning of buffer

	while ((long)ptr > 1 && (*ptr) && (long)ptr < (long)buffer + Data_Count) {
		if (*ptr == code) { return atof(ptr + 1); }								// convert the digits that follow into a float and return it
		ptr = strchr(ptr, ' ') + 1;												// take a step from here to the char. after the next space
	}
	return val;																	// If nothing found, return val.
}
/*
	Process commands.
*/
void processCommand() {

	char* ptr = buffer; // get start of buffer (see what the command starts with)

	if (*ptr == 'H' || *ptr == 'L' || *ptr == 'P' || *ptr == 'Q' || *ptr == 'S' || *ptr == 'V') {
		//	H
		/*
			Harmonic Gear Rotation
			Command is:
				H01 A<value> S<value> = Move to absolute degree +0 to 359, at a speed 0-255. (within 0.4 degree)
				H02 A<value> S<value> = Move to absolute degree -0 to 359, at a speed 0-255. (within 0.4 degree)
		*/
		if (cmd < 0) {
			cmd = ParseNumber('H', -1);
			switch (cmd) {
			case 1:  MoveGear(ParseNumber('A', Angle), true, ParseNumber('S', Current_Speed)); break;
			case 2:  MoveGear(ParseNumber('A', Angle), false, ParseNumber('S', Current_Speed)); break;
			default:
				break;
			}
		} // H

	}

}
/*
	Clear the Serial Buffer.
	Send "ok" handshake.
*/
void Clear_Buffer() {

	// init our command
	for (byte i = 0; i < MAX_BUF; i++) { buffer[i] = 0; }
	Data_Count = 0;
	ProcsessingCommand = false;
	isComment = false;
	cmd = -1;
	//Wire.flush();
	Serial.flush();
	Serial.println("ok");
	delay(4);
}
/*
	Read command in serial buffer.
*/
void ReadSerial() {

	char c;
	// read in characters if we got them.
	if (Serial.available() > 0) {
		c = Serial.read();
		no_data = 0;
		// newlines are ends of commands.
		if (c != '\r' || c != '\n') {
			if (c == '(' || c == ';') {
				isComment = true;
			}
			// If we're not in isComment mode, add it to our array.
			if (!isComment) {
				buffer[Data_Count++] = c;
			}
			if (c == ')') {
				isComment = false; // End of isComment - start listening again
			}
		}
	}
	else {
		no_data++;
		delayMicroseconds(100); // 100

		//	PROCESS CODE
		if (Data_Count && (c == '\n' || c == '\r' || no_data > 100)) {
			no_data = 0;
			Data_Count--;

#ifdef PRINT_SERIAL
			Serial.print("ReadSerial ");
			Serial.println(buffer);
			delay(50);
#endif	//	PRINT_SERIAL


			ProcsessingCommand = true;
			// cmd = -1;
			processCommand();      // Process Command
			Clear_Buffer(); // clear ommand.
		}
	}

}

/*	=====	Initialize	=====	*/

void setup() {
	/*
		Start Serial.
	*/
	Serial.begin(115200);
	delay(500);
	/*
		Start SPI Flash Filing System.
	*/
	if (SPIFFS.begin()) {
		Serial.println("SPIFFS Active\n");
		spiffsActive = true;
	}
	else {
		Serial.println("Unable to activate SPIFFS\n");
	}
	delay(500);
	/*
		INTERUPTS
	*/
	pinMode(SENS_A, INPUT);
	pinMode(SENS_B, INPUT);
	attachInterrupt(digitalPinToInterrupt(SENS_A), PinChange, CHANGE);
	attachInterrupt(digitalPinToInterrupt(SENS_B), PinChange, CHANGE);
	/*
		Pin mode.
	*/
	pinMode(MOTOR_SPEED, OUTPUT);
	pinMode(MOTOR_FORWARD, OUTPUT);
	pinMode(MOTOR_REVERSE, OUTPUT);
	/*
		Set Current_Speed 50%.
	*/
	SetSpeed(500);
	/*
		WiFi.
	*/
	WiFi.mode(WIFI_AP);
	WiFi.softAPConfig(local_ip, gateway, subnet);
#ifdef ACCESS_POINT
	/*
		Connect to ESP8266 Wi-Fi network with no password
	*/
	Serial.println("Setting AP (Access Point)");
	WiFi.softAP(ESP8266_SSID_AP);
	Serial.print("Connect you device to network '");
	Serial.print(ESP8266_SSID_AP);
	Serial.println("'\r\nThere is no Password.");
	Serial.print("Open browser to ");
	Serial.println(WiFi.softAPIP());

#endif // ACCESS_POINT
#ifdef ACCESS_POINT_PW
	/*
		Connect to ESP8266 Wi-Fi network with a password
	*/
	Serial.println("Setting AP (Access Point)");
	WiFi.softAP(ESP8266_SSID_AP, ESP8266_PASSWORD_AP);
	Serial.print("Connect you device to network '");
	Serial.print(ESP8266_SSID_AP);
	Serial.println("'");
	Serial.print("Use password '");
	Serial.print(ESP8266_PASSWORD_AP);
	Serial.println("'");
	Serial.print("Open browser to ");
	Serial.println(WiFi.softAPIP());

#endif // ACCESS_POINT_PW

	/*
		setup web server to handle specific HTTP requests.
	*/
	server.on("/", Handle_Root);
	server.on("/Front_View_IMAGE", HTTP_GET, Handle_Front_View_IMAGE);
	server.on("/Get_Speed", Handle_Get_Speed);
	server.on("/Get_Position", Handle_Get_Position);
	server.on("/0", HTTP_GET, Handle_0);
	server.on("/90", HTTP_GET, Handle_90);
	server.on("/180", HTTP_GET, Handle_180);
	server.on("/position", HTTP_GET, Handle_Position);
	server.on("/Set_Speed", HTTP_GET, Handle_Set_Speed);
	server.onNotFound(Handle_Not_Found);

	/*
		start server.
	*/


	server.begin();
	Serial.println("\nESP8266\nDrive Control\nserver started.\n");
	digitalWrite(LED_BUILTIN, LOW);


}

/*	=====	Loop	=====	*/

void loop() {
	/*
		 handle HTTP requests.
	*/
	server.handleClient();


	/*		=====	Serial Control	=====	*/
	/*
		Check if motor needs stoping.
	*/
	if (StopMotor) {
		StopMotor = false;
		MotorStop();

	}
	/*
		Check if motor needs slowing down.
	*/
	if (SlowMotor) {
		SlowMotor = false;
		SetSpeed(500);
	}
	/*
		Get next command when finish current command.
	*/
	if (!ProcsessingCommand) { ReadSerial(); }

	//Do_Test();
}
